@@ -31,6 +31,7 @@ Luke Arno can be found at http://lukearno.com/
 import urllib
 import cgi
 import mimetypes
+from cStringIO import StringIO
 
 from wsgiref import headers, util
 
@@ -94,16 +95,56 @@ class URI(object):
 class Request(object):
     """Yet another request object (for WSGI), as advertised."""
 
-    def __init__(self, environ):
+    def __init__(self, environ, start_response, extra_props=None):
         """Set up the various attributes."""
         self.environ = environ
+        self._start_response = start_response
+        self.start_response_called = False
         self.method = environ['REQUEST_METHOD']
         self.content_type = environ.get('CONTENT_TYPE', '')
         self.content_length = environ.get('CONTENT_LENGTH', '')
         self.uri = URI(environ)
         self.res = Response(self.uri())
         self._parse_query()
-        self._parse_form()
+        if extra_props is not None:
+            for prop_spec in extra_props:
+                self.load_prop(*prop_spec)
+        self.exc_info = None
+
+    def start_response(self, *a, **kw):
+        """Wrap the real start_response and set flag when called.
+        
+        This provides a means to prevent the real start_response from 
+        getting called more than once.
+
+        The flag is self.start_response_called = True|False.
+        """
+        self.start_response_called = True
+        return self._start_response(*a, **kw)
+
+    def load_prop(self, attr_name, key, default=None):
+        """Add an arbitrary property."""
+        if isinstance(key, str):
+            value = self.environ.get(key, default)
+        else:
+            value = key(self.environ)
+        setattr(self, attr_name, value)
+
+    def forward(self, app):
+        """Forward the request to another Yaro compatible handler."""
+        return app(self)
+
+    def wsgi_forward(self, wsgiapp):
+        """Forward the request to a WSGI compatible handler."""
+        return wsgiapp(self.environ, self.start_response)
+
+    def __getattr__(self, attr):
+        """Support lazy loading of form or body."""
+        if attr == 'form' and not 'form' in self.__dict__:
+            self._parse_form()
+        elif attr == 'body' and not 'body' in self.__dict__:
+            self._load_body()
+        return self.__dict__[attr]
 
     def _parse_query(self):
         """Populate a dictionary in self.query from the querystring.
@@ -157,30 +198,43 @@ class Request(object):
         If 'foo[]' and 'foo' are both used as input names, they 
         will step on each other. Use one or the other.
         """
-        fs = cgi.FieldStorage(fp=self.environ['wsgi.input'],
-                              environ=self.environ,
-                              keep_blank_values=True) 
         form = {}
-        for key in fs:
-            value = fs[key]
-            if key.endswith("[]"):
-                if not isinstance(value, list):
-                    value = [value]
-                key = key[:-2]
-            if isinstance(value, list):
-                newvalue = []
-                for v in value:
-                    if v.filename:
-                        newvalue.append(v)
-                    else:
-                        newvalue.append(v.value)
-                form[key] = newvalue
-            else:
-                if value.filename is None:
-                    value = value.value
-                form[key] = value
+        clen = self.environ.get('CONTENT_LENGTH', 0)
+        if clen:
+            sio = StringIO(self.environ['wsgi.input'].read(int(clen)))
+            fs = cgi.FieldStorage(fp=sio,
+                                  environ=self.environ,
+                                  keep_blank_values=True) 
+            sio.seek(0)
+            self.environ['wsgi.input'] = sio
+            for key in fs:
+                value = fs[key]
+                if key.endswith("[]"):
+                    if not isinstance(value, list):
+                        value = [value]
+                    key = key[:-2]
+                if isinstance(value, list):
+                    newvalue = []
+                    for v in value:
+                        if v.filename:
+                            newvalue.append(v)
+                        else:
+                            newvalue.append(v.value)
+                    form[key] = newvalue
+                else:
+                    if value.filename is None:
+                        value = value.value
+                    form[key] = value
         self.form = form
 
+    def _load_body(self):
+        """Set self.body with the raw body of the request."""
+        clen = self.environ.get('CONTENT_LENGTH', 0)
+        if clen:
+            self.body = self.environ['wsgi.input'].read(int(clen))
+        else:
+            self.body = ""
+
     def redirect(self, location, permanent=True):
         """Set the location header and status."""
         self.res.headers['Location'] = location
@@ -198,7 +252,7 @@ class Response(object):
         self._headers = []
         self.headers = headers.Headers(self._headers)
         guess = mimetypes.guess_type(uri)[0]
-        self.headers['Content-Type'] = guess or 'text/plain'
+        self.headers['Content-Type'] = guess or 'text/html'
         self.status = '200 OK'
         self.body = ""
 
@@ -206,17 +260,23 @@ class Response(object):
 class Yaro(object):
     """WSGI wrapper for something that takes and returns Request."""
 
-    def __init__(self, app):
+    def __init__(self, app, extra_props=None):
         """Take the thing to wrap."""
         self.app = app
+        self.extra_props = extra_props
         
     def __call__(self, environ, start_response):
         """Create Request, call thing, unwrap results and respond."""
-        environ['yaro.request'] = req = Request(environ)
+        if 'yaro.request' in environ:
+            req = environ['yaro.request']
+        else:
+            req = Request(environ, start_response, self.extra_props)
+            environ['yaro.request'] = req 
         body = self.app(req)
         if body is None:
             body = req.res.body
-        start_response(req.res.status, req.res._headers)
+        if not req.start_response_called:
+            req.start_response(req.res.status, req.res._headers, req.exc_info)
         if isinstance(body, str):
             return [body]
         elif isiterable(body):
@@ -244,11 +304,22 @@ if __name__ == '__main__':
     def foo(req):
         #req.res.body = 'Hello, World!'
         code.interact(local=locals(), banner="%s: %s" % (req.method, req.uri()))
-        if not req.uri.path.endswith('.txt'):
+        if req.uri.path.endswith('.rdr'):
             req.redirect('../proudhon.txt')
+        elif req.uri.path.endswith('.fwd'):
+            return req.forward(bar)
+        elif req.uri.path.endswith('.wsgi'):
+            return req.wsgi_forward(baz)
         else:
             req.res.body = req.uri()
 
+    def bar(req):
+        return "Hello Bar"
+
+    def baz(environ, start_response):
+        start_response("201 Created", [])
+        return ['Baz here.']
+
     try:
         make_server('localhost', 9999, Yaro(foo)).serve_forever()
     except KeyboardInterrupt, ki:
